/* 
 * CSE142, Spring 2000
 * Homework 5
 * Board Game

  STARTER PROGRAM

 */


#include "gp142.h"
#include <stdio.h>
#include <math.h>

#include <time.h>
	/* 
	* <time.h> is included so the current time can be accessed
	* to initialize the random number generator.  If you don't
	* use random numbers (and you don't have to) you can either
	* ignore this #include or delete it.
	*/

/* Logical constants */
#define TRUE    		1
#define FALSE   		0

/* 
 * For many of the GP142 functions, a line width of
 * zero means to fill the shape being drawn.  The
 * constant FILL is used in function calls when a solid
 * shape is desired.
 */
#define FILL 0


/*
 * Symbolic constants.  Change as needed.
 */

/* Menu buttons: */

/* Top-left coordinates of button palette */
#define MENU_TOP		( GP142_YMAX - 5)
#define MENU_LEFT		(-GP142_XMAX + 5)

/* Height and width of buttons in pixels */
#define BUTTON_HEIGHT  	25
#define BUTTON_WIDTH    75

/* Button-related constants */
#define NUM_BUTTONS  3     /* Number of buttons */
#define BUTTON_CLICK 1     /* Mouse click in some button */
#define OTHER_CLICK  0     /* Mouse click elsewhere (not on a button) */

/* Board-related constants */
#define BOARD_ROWS    5    /* number of rows of the board */
#define BOARD_COLS   10    /* number of columns of the board */
#define CELL_SIZE    ((2*GP142_XMAX) - (BUTTON_WIDTH) - 45) / (BOARD_COLS)
   /* cell size in screen coordinates (approx. 45) if size is 5 x 10 */

/* Player-related constants */
#define NUM_PLAYERS   1 /* the number of players in the game */
#define FIRST_MOVE    1 /* indicates that player has moved once and is about to jump if he/she falls on a cell with an arrow */
#define SECOND_MOVE   2 /* indicates that player has already jumped, so no further jumps for this turn are permitted */

/* Board arrow constants */
#define NO_ARROW         0 /* indicates a given cell has no arrow in it */
#define MOVE_LEFT        1 /* indicates a given cell has a left arrow in it (at most 1 arrow per cell!) */
#define MOVE_RIGHT       2 /* indicates a given cell has a right arrow in it */
#define MOVE_UP          3 /* indicates a given cell has an up arrow in it */
#define MOVE_DOWN        4 /* indicates a given cell has a down arrow in it */
#define NUM_LEFT_ARROWS  5 /* the number of left arrows to randomly place on the board */
#define NUM_RIGHT_ARROWS 5 /* the number of right arrows to randomly place on the board */
#define NUM_UP_ARROWS    5 /* the number of up arrows to randomly place on the board */
#define NUM_DOWN_ARROWS  5 /* the number of down arrows to randomly place on the board */


/* Structure definitions */
typedef struct
{
    int color; /* cell color */
	int arrow; /* type of arrow (if any) in a cell */
} cell; /* a structure storing data for a board cell */

typedef struct
{
	int color;       /* player color - to distinguish it from the rest of the players */
	int cell_number; /* the board cell where the player resides at a given moment */
	int status;      /* indicates whether a player has moved one or two times (the second one */
	                 /* with a jump) - related to FIRST_MOVE and SECOND_MOVE constants above */
} player; /* a structure, storing data for a player */

/*****************************************************************************

     FUNCTION PROTOTYPES

*****************************************************************************/

/* Initialize random number generator.  [Ignore or		*/
/* remove if your program doesn't use random numbers.]	*/
void initialize_random_numbers();

/* Draw buttons */
void draw_buttons(void);

/* Return kind of mouse click (in a button, etc.) */
int classify_mouse_click(int mouse_x, int mouse_y);

/* Process button click at location x, y. React appropriately, depending on the requested event */
void handle_button(int x, int y, int *die_value, cell board[BOARD_ROWS][BOARD_COLS], player *aPlayer);

/* Display error message if internal error is detected. */
void something_is_wrong(void);

/* Transform from a cell number between 1 and (BOARD_ROWS * BOARD_COLS) into board coordinates */
/* (i,j), where i is the row number and j - the column number                                  */
void transform_to_board_coord(int *i, int *j, int cell_number);

/* Initialize the board arrows */
void initialize_board(cell board[BOARD_ROWS][BOARD_COLS]);

/* Draws the game board with all associated arrows */
void draw_board(cell board[BOARD_ROWS][BOARD_COLS]);

/* Draw the die with its corresponding die value */
void draw_die(int die_value);

/* Update the die value by storing a random value between 1 and 6 into die_value */
void update_die(int *die_value);

/* Initialize player */
void initialize_player(player *aPlayer);

/* Draws a player */
void draw_player(player aPlayer);

/* Updates player position on the board based on the given die value. */
/* Note: A die value = 0 will not move the player by means of the die value itself, but if      */
/* the board contains an arrow in the cell where the player currently resides and if the player */
/* is about to make his second move, a jump in the appropriate direction (depending on the type */
/* of arrow) will be made. */
void update_player(cell board[BOARD_ROWS][BOARD_COLS], player *aPlayer, int die_value);


/*****************************************************************************

     MAIN

*****************************************************************************/

/*
 * Main program contents:
 *   - Major program variables that persist throughout execution
 *   - Event loop that processes GP142 events until the user quits,
 *     and updates variables and redraws the window as needed.
 */
int main(void)
{
	/* Event handling and buttons: */
	int  quit;				/* = "user has selected quit"				*/
	int  event;				/* the most recent GP142 event				*/
	char key_pressed;		/* last keyboard character from key event	*/
	int  mouse_x, mouse_y;	/* x and y coordinates of latest mouse event*/
	int  click_kind;		/* kind of mouse click (button, etc.)		*/

	player aPlayer;
    cell board[BOARD_ROWS][BOARD_COLS];
	int die_value = 1;

    /* Initialize random number generator */
	initialize_random_numbers();

	initialize_board(board);
	initialize_player(&aPlayer);
    draw_board(board);
	draw_die(die_value);
	draw_player(aPlayer);

	/* Initialize graphics package, clear the window,	*/
	/* turn logging off, and start animation.			*/
	GP142_open();
	GP142_logging(LOG_OFF);
	GP142_clear();
	GP142_animate(ANI_RUN);

	draw_buttons();
	
	/*
	 * Main event loop:
	 * ---- ----- -----
	 * Wait for the next user action, decode it, and call an
	 * appropriate function to handle it.  Repeat until the
	 * user selects quit.
	 */
	quit = FALSE;
	while (!quit) {
		
		/* get next event */
		event = GP142_await_event(&mouse_x, &mouse_y, &key_pressed);
		
		/* perform appropriate action depending on kind of event */
		switch (event) {
		case GP142_QUIT:
			/* Quit selected */
			quit = TRUE;
			break;
			
		case GP142_KBD:
			/* Key pressed */
            switch (key_pressed)
			{
			case ' ':
				update_die(&die_value);
				aPlayer.status = FIRST_MOVE;
				/* Move the player on its first move based on what he got on the die */
				update_player(board, &aPlayer, die_value);
				/* Move the player a second time (jump), by imitating a die of value 0.     */
				/* Such a die value will not move the player, unless the board has an arrow */
				/* in the cell where the player currently resides.                          */
				update_player(board, &aPlayer, 0);
				break;

			default:
				break;
			}
			break;
			
		case GP142_MOUSE:
			/* Mouse click.  Determine if in a menu button	*/
			/* or elsewhere and process accordingly.		*/
			click_kind = classify_mouse_click(mouse_x, mouse_y);

			if (click_kind == BUTTON_CLICK)
			{
				handle_button(mouse_x, mouse_y, &die_value, board, &aPlayer);
			}
			else
			{
				/* insert code here (and additional else-if statements)         */
				/* to react to mouse clicks elsewhere on the screen (if needed) */
			}
			break;
			
		case GP142_PERIODIC:
			/* Timer interval event.  Update everything	that needs the change */
			/* as time advances. Animation should go here!                    */
			break;
			
		default:
			/* unknown event */
			break;
			
		}  /*end switch */
		
		/* Redraw everything */
        draw_board(board);
		draw_die(die_value);
		draw_player(aPlayer);
		draw_buttons();
		
	}  /* end event loop */
	
	
	/* Termination: shut down graphics window and exit */
	GP142_close();
	return 0;
}  /* end main */


/*****************************************************************************

     INITIALIZATION PROCEDURES

*****************************************************************************/

/* Initialize random number generator */
void initialize_random_numbers()
{
	time_t t;			/* current time */
	srand (time(&t));
}

/* Initialize the board arrows */
void initialize_board(cell board[BOARD_ROWS][BOARD_COLS])
{
    int i, j, k;
	int cell_number;

	/* Initializes the cells as being all non-occupied, i.e. initially there are no arrows */
    for (i=0; i<BOARD_ROWS; i++)
		for (j=0; j<BOARD_COLS; j++)
		{
			board[i][j].arrow = NO_ARROW;
			board[i][j].color = TURQUOISE;
		}

	/* Randomly picks NUM_RIGHT_ARROWS different non-occupied cells and marks them    */
	/* as containing a MOVE_RIGHT arrow, where a rule to observe is that no arrow may */
	/* be placed on the very last (final) cell.                                       */
    k=0;
	while (k<NUM_RIGHT_ARROWS)
	{
		cell_number = rand() % (BOARD_ROWS * BOARD_COLS) + 1;

		/* Converting cell numbers into board coordinates. */
		transform_to_board_coord(&i, &j, cell_number);
		if (cell_number != (BOARD_ROWS * BOARD_COLS) &&
			board[i][j].arrow == NO_ARROW)
		{
            board[i][j].arrow = MOVE_RIGHT;
			k++;
		}
	}

	/* Put other arrows for MOVE_LEFT, MOVE_UP, MOVE_DOWN here */
    /* ... */
}

/* Initialize player */
void initialize_player(player *aPlayer)
{
	aPlayer->cell_number = 0;
	aPlayer->color = PEACH;
	aPlayer->status = FIRST_MOVE;
}


/*****************************************************************************

     BUTTONS AND MOUSE CLICKS

*****************************************************************************/

/* Classify mouse click at (mouse_x,mouse_y) and return kind */
int classify_mouse_click(int mouse_x, int mouse_y)
{
	if ((mouse_x > MENU_LEFT) &&
		(mouse_x < MENU_LEFT + BUTTON_WIDTH) &&
		(mouse_y < MENU_TOP) &&
		(mouse_y > (MENU_TOP - NUM_BUTTONS * BUTTON_HEIGHT)))
		return BUTTON_CLICK;
	else
		return OTHER_CLICK;
}

/* Process button click at location x, y. React appropriately: */
/* i.e. change colors, tree tops, etc. if requested.           */
void handle_button(int x, int y, int *die_value, cell board[BOARD_ROWS][BOARD_COLS], player *aPlayer)
{
	int buttonNum; /* number of selected button */

	/* Complain if x coordinate couldn't be inside a button */
	if (x < MENU_LEFT || x > MENU_LEFT + BUTTON_WIDTH) {
		something_is_wrong();
		return;
	}

	/* Convert y coordinate to button number from top. */
	buttonNum = (MENU_TOP - y)/BUTTON_HEIGHT;	 
	
	switch(buttonNum) {
	case 0:
		/* Change the board arrows */
	    initialize_board(board);
		break;
	case 1:
		/* Roll the die */
        update_die(die_value);
		aPlayer->status = FIRST_MOVE;
		/* Move the player on its first move based on what he got on the die */
		update_player(board, aPlayer, *die_value);
    	/* Move the player a second time (jump), by imitating a die of value 0.    */
		/* Such a die value will not result in a player move, unless the board has */
		/* an arrow in the cell where the player currently resides.                */
		update_player(board, aPlayer, 0);
		break;
		
	case 2:
		/* Put code to restart game */
		/* ... */

		break;

	default:
		/*Bad news if this is ever reached! */
	    GP142_animate(ANI_HALT);
		something_is_wrong();
		break;
	}
}


/*****************************************************************************

     DRAWING PROCEDURES

*****************************************************************************/

/* Draws the game board with all associated arrows */
void draw_board(cell board[BOARD_ROWS][BOARD_COLS])
{
    int i, j, cell_number;
	int x0, y0, x1, y1;
	int x = GP142_XMAX - 25;
	int y = GP142_YMAX - 25;
	int board_height = BOARD_ROWS * CELL_SIZE;
	int board_width  = BOARD_COLS * CELL_SIZE;

    /* Draw board outline */
	GP142_rectangleXY(board[0][0].color, x, y, x - board_width, y - board_height, FILL);

    /* Draw cell boundaries */
	for (i=0; i<=BOARD_COLS; i++)
	{
		GP142_lineXY(BLACK, x, y, x, y - board_height, 2);
        x = x - CELL_SIZE;
	}
	x = GP142_XMAX - 25;
	for (i=0; i<=BOARD_ROWS; i++)
	{
		GP142_lineXY(BLACK, x, y, x - board_width, y, 2);
        y = y - CELL_SIZE;
	}

	/* Illustrate arrows by drawing arrows in corresponding cells */
    for (i=0; i<BOARD_ROWS; i++)
		for (j=0; j<BOARD_COLS; j++)
		{
			/* Print cell numbers in lower left corners */
            cell_number = (BOARD_ROWS - i - 1) * BOARD_COLS + j + 1;
			x0 = GP142_XMAX - 25 - BOARD_COLS * CELL_SIZE + j * CELL_SIZE + CELL_SIZE / 12;
			y0 = GP142_YMAX - 25 - (i+1) * CELL_SIZE + CELL_SIZE / 12;
			GP142_printfXY(BLACK, x0, y0, 10, "%d", cell_number);

			/* Draw arrows */
			switch (board[i][j].arrow)
			{
			case NO_ARROW: /* Nothing to be drawn */
				break;

			case MOVE_LEFT:
                /* Put your code here */
				break;

			case MOVE_RIGHT:
				/* Draws a right arrow in the upper portion of the cell with coordinates (i,j) */
				x0 = (GP142_XMAX - 25 - BOARD_COLS * CELL_SIZE) + (j * CELL_SIZE) + (CELL_SIZE / 6);
				y0 = (GP142_YMAX - 25) - (i * CELL_SIZE) - (CELL_SIZE / 6);
				x1 = x0 + 2 * CELL_SIZE / 3;
				y1 = y0;
                GP142_lineXY(RED, x0, y0, x1, y1, 2);
				GP142_lineXY(RED, x1 - CELL_SIZE / 4, y1 + CELL_SIZE / 12, x1, y1, 2);
				GP142_lineXY(RED, x1 - CELL_SIZE / 4, y1 - CELL_SIZE / 12, x1, y1, 2);
				break;

			case MOVE_UP:
                /* Put your code here */
				break;

			case MOVE_DOWN:
                /* Put your code here */
				break;

			default:
				GP142_animate(ANI_HALT);
				something_is_wrong();
			}
		}
}

/* Draw a die with its corresponding die value */
void draw_die(int die_value)
{
	int x0, y0, x1, y1;
	x0 = 0;
	y0 = GP142_YMAX - 75 - BOARD_ROWS * CELL_SIZE;
	x1 = CELL_SIZE;
	y1 = GP142_YMAX - 75 - BOARD_ROWS * CELL_SIZE - CELL_SIZE;
    GP142_rectangleXY(FOREST_GREEN, x0, y0, x1, y1, 2);
	x1 = x0 + CELL_SIZE / 2;
	y1 = y0 + CELL_SIZE / 4;
	GP142_lineXY(FOREST_GREEN, x0, y0, x1, y1, 2);
	x0 = x1;
	y0 = y1;
	x1 = x0 + CELL_SIZE;
    /* y1 stays the same here */
    GP142_lineXY(FOREST_GREEN, x0, y0, x1, y1, 2);
	x0 = x1;
    y0 = y1;
	/* x1 stays the same */
	y1 = y1 - CELL_SIZE;
	GP142_lineXY(FOREST_GREEN, x0, y0, x1, y1, 2);
    x0 = x1;
	y0 = y1;
	x1 = CELL_SIZE;
	y1 = GP142_YMAX - 75 - BOARD_ROWS * CELL_SIZE - CELL_SIZE;
	GP142_lineXY(FOREST_GREEN, x0, y0, x1, y1, 2);
    x0 = CELL_SIZE;
	y0 = GP142_YMAX - 75 - BOARD_ROWS * CELL_SIZE;
    x1 = x0 + CELL_SIZE / 2;
	y1 = y0 + CELL_SIZE / 4;
	GP142_lineXY(FOREST_GREEN, x0, y0, x1, y1, 2);
	/* "Delete" the previous die value by overwriting it with a rectangle of the background color */
	GP142_rectangleXY(WHITE, CELL_SIZE / 10, y0 - CELL_SIZE / 10, 9 * CELL_SIZE / 10, y0 - 9 * CELL_SIZE / 10, 0);
	/* Put in the new die value */
	GP142_printfXY(NAVY_BLUE, CELL_SIZE / 3, y0 - 3 * CELL_SIZE / 4, 24, "%d", die_value);
}

/* Draws a player of appropriate color and position (as indicated in the player structure) */
void draw_player(player aPlayer)
{
	int x, y;
	int i, j;

	if (aPlayer.cell_number == 0)
	{
		/* Player not on the board yet */
		x = GP142_XMAX - 25 - BOARD_COLS * CELL_SIZE - CELL_SIZE / 2;
		y = GP142_YMAX - 25 - BOARD_ROWS * CELL_SIZE + CELL_SIZE / 2;
	}
	else
	{
		/* Convert from cell_number to board coordinates (i,j) */
        transform_to_board_coord(&i, &j, aPlayer.cell_number);
    	x = GP142_XMAX - 25 - BOARD_COLS * CELL_SIZE + j * CELL_SIZE + CELL_SIZE / 2;
	    y = GP142_YMAX - 25 - (i+1) * CELL_SIZE + CELL_SIZE / 2;
	}

	/* Delete start out-of-board player position by overwriting with the background color */
    GP142_circleXY(WHITE,
		GP142_XMAX - 25 - BOARD_COLS * CELL_SIZE - CELL_SIZE / 2,
		GP142_YMAX - 25 - BOARD_ROWS * CELL_SIZE + CELL_SIZE / 2, CELL_SIZE / 6);

	/* Update player position */
    GP142_circleXY(aPlayer.color, x, y, CELL_SIZE / 6);
}

/* Draws the buttons */
void draw_buttons(void)
{
	int i;
	int text_offset = MENU_LEFT+10;	/* offset of button text in button */
	int font_size = 12;				/* size of button text */
	
	/* draw rectangles for each button */
	for(i = 0; i < NUM_BUTTONS; i++) {
		GP142_rectangleXY(BLACK,
			MENU_LEFT,
			MENU_TOP - i * BUTTON_HEIGHT,
			MENU_LEFT + BUTTON_WIDTH,
			MENU_TOP - (i + 1) * BUTTON_HEIGHT, 1);
	}
	
	/* display button text */
	i = MENU_TOP - (BUTTON_HEIGHT/2) - (font_size/2);
	GP142_printfXY(BLACK, text_offset, i, font_size, "Alter arrows");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(BLACK, text_offset, i, font_size, "Roll die");
	
	i -= BUTTON_HEIGHT;
	GP142_printfXY(BLACK, text_offset, i, font_size, "Restart");
}


/*****************************************************************************

     OTHER

*****************************************************************************/

/* Transform from a cell number between 1 and (BOARD_ROWS * BOARD_COLS) into board coordinates */
/* (i,j), where i is the row number and j - the column number                                  */
void transform_to_board_coord(int *i, int *j, int cell_number)
{
    *i = BOARD_ROWS - 1 - (cell_number - 1) / BOARD_COLS;
	*j = (cell_number - 1) % BOARD_COLS;
}

/* Update the die value	with a random number between 1 and 6 */
void update_die(int *die_value)
{
	*die_value = rand() % 6 + 1;
}

/* Updates player position on the board based on the given die value. */
/* Note: A die value = 0 will not move the player by means of the die value itself, but if      */
/* the board contains an arrow in the cell where the player currently resides and if the player */
/* is about to make his second move, a jump in the appropriate direction (depending on the type */
/* of arrow) will be made. */
void update_player(cell board[BOARD_ROWS][BOARD_COLS], player *aPlayer, int die_value)
{
	int i, j; /* board coordinates (i,j) */

    /* Update the player position, based on the die value */
	aPlayer->cell_number = aPlayer->cell_number + die_value;
	/* If player reached the goal - the last cell on the board, he becomes a winner! */
	if (aPlayer->cell_number == BOARD_COLS * BOARD_ROWS)
	{
		/* Congratulate the winner and quit the procedure */
		/* ... */
	}
	else
	{
		/* Otherwise, if the player was about to jump overboard, he stays in place */
		if (aPlayer->cell_number > BOARD_COLS * BOARD_ROWS)
		    aPlayer->cell_number = aPlayer->cell_number - die_value;

	    /* Convert from cell_number to board coordinates (i,j) */
	    transform_to_board_coord(&i, &j, aPlayer->cell_number);
	    /* If there is an arrow in cell (i,j) of the board and the player has moved only once... */
	    if (board[i][j].arrow != NO_ARROW && aPlayer->status == FIRST_MOVE)
		{
		    /* ... a jump follows, according to the type of arrow the cell contains */
            aPlayer->status = SECOND_MOVE;
            switch (board[i][j].arrow)
			{
		    case MOVE_LEFT:
			    /* Put in the appropriate code here */
			    /* ... */
			    break;

		    case MOVE_RIGHT:
				/* the player's new position will be in the next cell (with number one higher) */
			    aPlayer->cell_number++;
			    break;

    	    case MOVE_UP:
			    /* Put in the appropriate code here */
			    /* ... */
			    break;

		    case MOVE_DOWN:
			    /* Put in the appropriate code here */
			    /* ... */
			    break;

		    default:
			    GP142_animate(ANI_HALT);
			    something_is_wrong();
			    break;
			}
		}
	}
}

/* Display error message if internal error is detected. */
void something_is_wrong (void) {
	GP142_printfXY (RED, 0, 0, 16, "Program error!");
}